from clikit.api.io import IO
from clikit.api.config import ApplicationConfig
from clikit.ui import Component

from .paragraph import Paragraph


class NameVersion(Component):
    """
    Renders the name and version of an application.
    """

    def __init__(self, config):  # type: (ApplicationConfig) -> None
        self._config = config

    def render(self, io, indentation=0):  # type: (IO, int) -> None
        if self._config.display_name and self._config.version:
            paragraph = Paragraph(
                "{} version <c1>{}</c1>".format(
                    self._config.display_name, self._config.version
                )
            )
        elif self._config.display_name:
            paragraph = Paragraph("{}".format(self._config.display_name))
        else:
            paragraph = Paragraph("Console Tool")

        paragraph.render(io, indentation)



import os
import re

from .question import Question


class SelectChoiceValidator:
    def __init__(self, question):
        """
        Constructor.
        """
        self._question = question
        self._values = question.choices

    def validate(self, selected):
        """
        Validate a choice.
        """
        # Collapse all spaces.
        if isinstance(selected, int):
            selected = str(selected)

        selected_choices = selected.replace(" ", "")

        if self._question.supports_multiple_choices():
            # Check for a separated comma values
            if not re.match("^[a-zA-Z0-9_-]+(?:,[a-zA-Z0-9_-]+)*$", selected_choices):
                raise ValueError(self._question.error_message.format(selected))

            selected_choices = selected_choices.split(",")
        else:
            selected_choices = [selected]

        multiselect_choices = []
        for value in selected_choices:
            results = []

            for key, choice in enumerate(self._values):
                if choice == value:
                    results.append(key)

            if len(results) > 1:
                raise ValueError(
                    "The provided answer is ambiguous. Value should be one of {}.".format(
                        " or ".join(str(r) for r in results)
                    )
                )

            try:
                result = self._values.index(value)
                result = self._values[result]
            except ValueError:
                try:
                    value = int(value)

                    if value < len(self._values):
                        result = self._values[value]
                    else:
                        result = False
                except ValueError:
                    result = False

            if result is False:
                raise ValueError(self._question.error_message.format(value))

            multiselect_choices.append(result)

        if self._question.supports_multiple_choices():
            return multiselect_choices

        return multiselect_choices[0]


class ChoiceQuestion(Question):
    """
    Multiple choice question.
    """

    def __init__(self, question, choices, default=None):
        super(ChoiceQuestion, self).__init__(question, default)

        self._multi_select = False
        self._choices = choices
        self._validator = SelectChoiceValidator(self).validate
        self._autocomplete_values = choices
        self._prompt = " > "
        self._error_message = 'Value "{}" is invalid'

    @property
    def error_message(self):
        return self._error_message

    @property
    def choices(self):
        return self._choices

    def supports_multiple_choices(self):
        return self._multi_select

    def set_multi_select(self, multi_select):
        self._multi_select = multi_select

    def set_error_message(self, message):
        self._error_message = message

    def _write_prompt(self, io):
        """
        Outputs the question prompt.
        """
        message = self._question
        default = self._default

        if default is None:
            message = "<question>{}</>: ".format(message)
        elif self._multi_select:
            choices = self._choices
            default = default.split(",")

            for i, value in enumerate(default):
                default[i] = choices[int(value.strip())]

            message = "<question>{}</> [<comment>{}</>]:".format(
                message, ", ".join(default)
            )
        else:
            choices = self._choices
            message = "<question>{}</question> [<comment>{}</>]:".format(
                message, choices[int(default)]
            )

        if len(self._choices) > 1:
            width = max(*map(len, [str(k) for k, _ in enumerate(self._choices)]))
        else:
            width = 1

        messages = [message]
        for key, value in enumerate(self._choices):
            messages.append(" [<comment>{:{}}</>] {}".format(key, width, value))

        io.error_line("\n".join(messages))

        message = self._prompt

        io.error(message)



from contextlib import contextmanager

from clikit.api.io import IO
from clikit.ui import Component
from clikit.ui.alignment import LabelAlignment
from clikit.ui.components import LabeledParagraph


class BlockLayout:
    """
    Renders renderable objects in indented blocks.
    """

    def __init__(self):  # type: () -> None
        self._current_indentation = 0
        self._elements = []
        self._indentations = []
        self._alignment = LabelAlignment()

    def add(self, element):  # type: (Component) -> BlockLayout
        self._elements.append(element)
        self._indentations.append(self._current_indentation)

        if isinstance(element, LabeledParagraph):
            self._alignment.add(element, self._current_indentation)
            element.set_alignment(self._alignment)

        return self

    @contextmanager
    def block(self):  # type: () -> BlockLayout
        self._current_indentation += 2

        yield self

        self._current_indentation -= 2

    def render(self, io, indentation=0):  # type: (IO, int) -> None
        self._alignment.align(io, indentation)

        for i, element in enumerate(self._elements):
            element.render(io, self._indentations[i] + indentation)

        self._elements = []



import sys

from .api.application import Application as BaseApplication
from .api.args.raw_args import RawArgs
from .api.args.format.args_format import ArgsFormat
from .api.command import Command
from .api.command import CommandCollection
from .api.command.exceptions import CannotAddCommandException
from .api.config.application_config import ApplicationConfig
from .api.config.command_config import CommandConfig
from .api.event import CONFIG
from .api.event import PRE_RESOLVE
from .api.event import ConfigEvent
from .api.event import PreResolveEvent
from .api.exceptions import CliKitException
from .api.io import IO
from .api.io import InputStream
from .api.io import OutputStream
from .api.io.flags import VERY_VERBOSE
from .api.resolver.resolved_command import ResolvedCommand
from .args.argv_args import ArgvArgs
from .io import ConsoleIO
from .ui.components.exception_trace import ExceptionTrace


class ConsoleApplication(BaseApplication):
    """
    A console application
    """

    def __init__(self, config):  # type: (ApplicationConfig) -> None
        self._preliminary_io = ConsoleIO()

        # Enable trace output for exceptions thrown during boot
        self._preliminary_io.set_verbosity(VERY_VERBOSE)

        self._dispatcher = None

        try:
            dispatcher = config.dispatcher
            if dispatcher and dispatcher.has_listeners(CONFIG):
                dispatcher.dispatch(CONFIG, ConfigEvent(config))

            self._config = config
            self._dispatcher = config.dispatcher
            self._commands = CommandCollection()
            self._named_commands = CommandCollection()
            self._default_commands = CommandCollection()
            self._global_args_format = ArgsFormat(
                list(config.arguments.values()) + list(config.options.values())
            )

            for command_config in config.command_configs:
                self.add_command(command_config)
        except Exception as e:
            if not config.is_exception_caught():
                raise

            # Render the trace to the preliminary IO
            trace = ExceptionTrace(e)
            trace.render(self._preliminary_io)

            # Ignore is_terminated_after_run() setting. This is a fatal error.
            sys.exit(self.exception_to_exit_code(e))

    @property
    def config(self):  # type: () -> ApplicationConfig
        return self._config

    @property
    def global_args_format(self):  # type: () -> ArgsFormat
        return self._global_args_format

    def get_command(self, name):  # type: (str) -> Command
        return self._commands.get(name)

    @property
    def commands(self):  # type: () -> CommandCollection
        return self._commands

    def has_command(self, name):  # type: (str) -> bool
        return name in self._commands

    def has_commands(self):  # type: () -> bool
        return not self._commands.is_empty()

    @property
    def named_commands(self):  # type: () -> CommandCollection
        return self._named_commands

    def has_named_commands(self):  # type: () -> bool
        return not self._named_commands.is_empty()

    @property
    def default_commands(self):  # type: () -> CommandCollection
        return self._default_commands

    def has_default_commands(self):  # type: () -> bool
        return not self._default_commands.is_empty()

    def resolve_command(self, args):  # type: (RawArgs) -> ResolvedCommand
        if self._dispatcher and self._dispatcher.has_listeners(PRE_RESOLVE):
            event = PreResolveEvent(args, self)
            self._dispatcher.dispatch(PRE_RESOLVE, event)

            resolved_command = event.resolved_command
            if resolved_command:
                return resolved_command

        return self._config.command_resolver.resolve(args, self)

    def run(
        self, args=None, input_stream=None, output_stream=None, error_stream=None
    ):  # type: (RawArgs, InputStream, OutputStream, OutputStream) -> int
        # Render errors to the preliminary IO until the final IO is created
        io = self._preliminary_io
        try:
            if args is None:
                args = ArgvArgs()

            io_factory = self._config.io_factory

            io = io_factory(
                self, args, input_stream, output_stream, error_stream
            )  # type: IO

            resolved_command = self.resolve_command(args)
            command = resolved_command.command
            parsed_args = resolved_command.args

            status_code = command.handle(parsed_args, io)
        except KeyboardInterrupt:
            status_code = 1
        except Exception as e:
            if not self._config.is_exception_caught():
                raise

            trace = ExceptionTrace(e)
            trace.render(io, simple=isinstance(e, CliKitException))

            status_code = self.exception_to_exit_code(e)

        if self._config.is_terminated_after_run():
            sys.exit(status_code)

        return status_code

    def exception_to_exit_code(self, e):  # type: (Exception) -> int
        if not hasattr(e, "code") or not isinstance(e, int):
            return 1

        return min(max(e.code, 1), 255)

    def add_command(self, config):  # type: (CommandConfig) -> None
        if not config.is_enabled():
            return

        self._validate_command_name(config.name)

        command = Command(config, self)
        self._commands.add(command)

        if config.is_default():
            self._default_commands.add(command)

        if not config.is_anonymous():
            self._named_commands.add(command)

    def _validate_command_name(self, name):  # type: (Optional[str]) -> None
        if not name:
            raise CannotAddCommandException.name_empty()

        if name in self._commands:
            raise CannotAddCommandException.name_exists(name)



from typing import List
from typing import Optional


class RawArgs(object):
    """
    The unparsed console arguments.
    """

    @property
    def script_name(self):  # type: () -> Optional[str]
        raise NotImplementedError()

    @property
    def tokens(self):  # type: () -> List[str]
        raise NotImplementedError()

    def has_token(self, token):  # type: (str) -> bool
        raise NotImplementedError()

    def to_string(self, script_name=True):  # type: (bool) -> str
        raise NotImplementedError()

    @property
    def option_tokens(self):  # type: () -> List[str]
        raise NotImplementedError()
        return list(itertools.takewhile(lambda arg: arg != "--", self.tokens))

    def has_option_token(self, token):  # type: (str) -> bool
        raise NotImplementedError()



class Event(object):
    """
    Event
    """

    def __init__(self):  # type: () -> None
        self._propagation_stopped = False

    def is_propagation_stopped(self):  # type: () -> bool
        return self._propagation_stopped

    def stop_propagation(self):  # type: () -> None
        self._propagation_stopped = True



from clikit.api.args import Args
from clikit.api.args import RawArgs
from clikit.api.args.exceptions import CannotParseArgsException
from clikit.api.command import Command


class ResolveResult(object):
    """
    An intermediate result created during resolving.
    """

    def __init__(self, command, raw_args):  # type: (Command, RawArgs) -> None
        self._command = command
        self._raw_args = raw_args
        self._parsed_args = None
        self._parse_error = None
        self._parsed = False

    @property
    def command(self):  # type: () -> Command
        return self._command

    @property
    def raw_args(self):  # type: () -> RawArgs
        return self._raw_args

    @property
    def parsed_args(self):  # type: () -> Args
        if not self._parsed:
            self._parse()

        return self._parsed_args

    @property
    def parse_error(self):  # type: () -> CannotParseArgsException
        if not self._parsed:
            self._parse()

        return self._parse_error

    def is_parsable(self):  # type: () -> bool
        if not self._parsed:
            self._parse()

        return self.parse_error is None

    def _parse(self):  # type: () -> None
        try:
            self._parsed_args = self._command.parse(self._raw_args)
        except CannotParseArgsException as e:
            self._parse_error = e

        self._parsed = True



from clikit.api.args import Args
from clikit.api.command import Command


class ResolvedCommand(object):
    """
    A resolved command.
    """

    def __init__(self, command, args):  # type: (Command, Args) -> None
        self._command = command
        self._args = args

    @property
    def command(self):  # type: () -> Command
        return self._command

    @property
    def args(self):  # type: () -> Args
        return self._args



from typing import Dict
from typing import List
from typing import Optional

from .style import Style


class StyleSet(object):
    """
    A set of styles used by the formatter.
    """

    def __init__(self, styles=None):  # type: (Optional[List[Style]]) -> None
        if styles is None:
            styles = []

        self._styles = {}

        for style in styles:
            self.add(style)

    @property
    def styles(self):  # type: () -> (Dict[str, Style])
        return self._styles

    def add(self, style):  # type: (Style) -> None
        if not style.tag:
            raise ValueError("The tag of a style added to the style set must be set.")

        self._styles[style.tag] = style

    def replace(self, styles):  # type: (Optional[List[Style]]) -> None
        self._styles = {}

        for style in styles:
            self.add(style)

    def remove(self, tag):  # type: (str) -> None
        if tag in self._styles:
            del self._styles[tag]



from clikit.api.args.args import Args
from clikit.api.io import IO


class HelpHandler:
    """
    Handler for the "help" command.
    """

    def handle(self, args, io):  # type: (Args, IO) -> int
        print(args)

        return 0



from clikit.api.args import Args
from clikit.api.command import Command
from clikit.api.io import IO
from clikit.api.resolver import CommandResolver

from clikit.ui.help import ApplicationHelp
from clikit.ui.help import CommandHelp


class HelpTextHandler:
    """
    Displays help as text format.
    """

    def __init__(self, resolver):  # type: (CommandResolver) -> None
        self._resolver = resolver

    def handle(self, args, io, command):  # type: (Args, IO, Command) -> int
        application = command.application

        if args.is_argument_set("command"):
            resolved_command = self._resolver.resolve(args.raw_args, application)
            the_command = resolved_command.command

            usage = CommandHelp(the_command)
        else:
            usage = ApplicationHelp(application)

        usage.render(io)

        return 0



from clikit.api.args.raw_args import RawArgs
from clikit.api.resolver import ResolvedCommand

from .default_resolver import DefaultResolver
from .resolve_result import ResolveResult


class HelpResolver(DefaultResolver):
    """
    A command resolver used by a help handler.
    """

    def __init__(self, help_command_name="help"):  # type: (str) -> None:
        self._help_command_name = help_command_name

    def resolve(
        self, args, application
    ):  # type: (RawArgs, Application) -> ResolvedCommand
        if args.tokens and args.tokens[0] == self._help_command_name:
            del args.tokens[0]

        return super(HelpResolver, self).resolve(args, application)

    def create_resolved_command(
        self, result
    ):  # type: (ResolveResult) -> ResolvedCommand
        result.command.config.enable_lenient_args_parsing()

        resolved_command = super(HelpResolver, self).create_resolved_command(result)

        result.command.config.disable_lenient_args_parsing()

        return resolved_command



from typing import List

from clikit.utils._compat import OrderedDict

from .command import Command
from .exceptions import NoSuchCommandException


class CommandCollection(object):
    """
    A collection of named commands.
    """

    def __init__(self, commands=None):  # type: (List[Command]) -> None
        if commands is None:
            commands = []

        self._commands = OrderedDict()
        self._short_name_index = OrderedDict()
        self._alias_index = OrderedDict()

        for command in commands:
            self.add(command)

    def add(self, command):  # type: (Command) -> CommandCollection
        name = command.name

        self._commands[name] = command

        short_name = command.short_name
        if short_name:
            self._short_name_index[short_name] = name

        for alias in command.aliases:
            self._alias_index[alias] = name

        return self

    def get(self, name):  # type: (str) -> Command
        if name in self._commands:
            return self._commands[name]

        if name in self._short_name_index:
            return self._commands[self._short_name_index[name]]

        if name in self._alias_index:
            return self._commands[self._alias_index[name]]

        raise NoSuchCommandException(name)

    def is_empty(self):  # type: () -> bool
        return not self._commands

    def get_names(self, include_aliases=False):  # type: (bool) -> List[str]
        names = list(self._commands.keys())

        if include_aliases:
            names += list(self._alias_index.keys())

        names.sort()

        return names

    def __contains__(self, name):
        return (
            name in self._commands
            or name in self._short_name_index
            or name in self._alias_index
        )

    def __iter__(self):
        return iter(self._commands.values())

    def __len__(self):
        return len(self._commands)



from .config_event import ConfigEvent
from .console_events import CONFIG
from .console_events import PRE_HANDLE
from .console_events import PRE_RESOLVE
from .event import Event
from .event_dispatcher import EventDispatcher
from .pre_handle_event import PreHandleEvent
from .pre_resolve_event import PreResolveEvent



from typing import List
from typing import Optional

from clikit.api.args.args import Args
from clikit.api.args.raw_args import RawArgs
from clikit.api.args.format.args_format import ArgsFormat
from clikit.api.config.command_config import CommandConfig
from clikit.api.event import PRE_HANDLE
from clikit.api.event import PreHandleEvent
from clikit.api.io import IO


class Command(object):
    """
    A console command.
    """

    def __init__(
        self, config, application=None, parent_command=None
    ):  # type: (CommandConfig, Optional[Application], Optional[Command]) -> None
        from .command_collection import CommandCollection

        if not config.name:
            raise RuntimeError("The name of the command config must be set.")

        self._name = config.name
        self._short_name = None
        self._aliases = config.aliases
        self._config = config
        self._application = application
        self._parent_command = parent_command
        self._sub_commands = CommandCollection()
        self._named_sub_commands = CommandCollection()
        self._default_sub_commands = CommandCollection()
        self._args_format = config.build_args_format(self.base_format)
        self._dispatcher = application.config.dispatcher if application else None

        for sub_config in config.sub_command_configs:
            self.add_sub_command(sub_config)

    @property
    def name(self):  # type: () -> str
        return self._name

    @property
    def short_name(self):  # type: () -> str
        return self._short_name

    @property
    def full_name(self):  # type: () -> str
        if self._parent_command:
            return "{} {}".format(str(self._parent_command.full_name), self._name)

        return self._name

    @property
    def aliases(self):  # type: () -> List[str]
        return self._aliases

    def has_aliases(self):  # type: () -> bool
        return len(self._aliases) > 0

    @property
    def config(self):  # type: () -> CommandConfig
        return self._config

    @property
    def args_format(self):  # type: () -> ArgsFormat
        return self._args_format

    @property
    def application(self):  # type: () -> Application
        return self._application

    @property
    def parent_command(self):  # type: () -> Command
        return self._parent_command

    @property
    def sub_commands(self):  # type: () -> CommandCollection
        return self._sub_commands

    def get_sub_command(self, name):  # type: (str) -> Command
        return self._sub_commands.get(name)

    def has_sub_commands(self):  # type: () -> bool
        return len(self._sub_commands) > 0

    @property
    def named_sub_commands(self):  # type: () -> CommandCollection
        return self._named_sub_commands

    def get_named_sub_command(self, name):  # type: (str) -> Command
        return self._named_sub_commands.get(name)

    def has_named_sub_commands(self):  # type: () -> bool
        return len(self._named_sub_commands) > 0

    @property
    def default_sub_commands(self):  # type: () -> CommandCollection
        return self._default_sub_commands

    def get_default_sub_command(self, name):  # type: (str) -> Command
        return self._default_sub_commands.get(name)

    def has_default_sub_commands(self):  # type: () -> bool
        return len(self._default_sub_commands) > 0

    def parse(self, args, lenient=None):  # type: (RawArgs, Optional[bool]) -> Args
        if lenient is None:
            lenient = self._config.is_lenient_args_parsing_enabled()

        return self._config.args_parser.parse(args, self._args_format, lenient)

    def run(self, args, io):  # type: (RawArgs, IO) -> int
        return self.handle(self.parse(args), io)

    def handle(self, args, io):  # type: (Args, IO) -> int
        try:
            status_code = self._do_handle(args, io)
        except KeyboardInterrupt:
            if io.is_debug():
                raise

            status_code = 1

        # Any empty value is considered a success
        if not status_code:
            return 0

        # Anything else is normalized to a valid error status code
        return min(max(int(status_code), 1), 255)

    @property
    def base_format(self):  # type: () -> Optional[ArgsFormat]
        if self._parent_command:
            return self._parent_command.args_format

        if self._application:
            return self._application.global_args_format

        return

    def add_sub_command(self, config):  # type: (CommandConfig) -> None
        if not config.is_enabled():
            return

        command = self.__class__(config, self._application, self)

        # TODO: Validate command

        self._sub_commands.add(command)

        if config.is_default():
            self._default_sub_commands.add(command)

        if not config.is_anonymous():
            self._named_sub_commands.add(command)

    def _do_handle(self, args, io):  # type: (Args, IO) -> Optional[int]
        if self._dispatcher and self._dispatcher.has_listeners(PRE_HANDLE):
            event = PreHandleEvent(args, io, self)
            self._dispatcher.dispatch(PRE_HANDLE, event)

            if event.is_handled():
                return event.status_code

        handler = self._config.handler
        handler_method = self._config.handler_method

        return getattr(handler, handler_method)(args, io, self)

    def __repr__(self):  # type: () -> str
        return "<Command {}>".format(self.full_name)



from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from clikit.utils._compat import OrderedDict

from ..exceptions import CannotAddArgumentException
from ..exceptions import CannotAddOptionException
from ..exceptions import NoSuchArgumentException
from ..exceptions import NoSuchOptionException
from .args_format import ArgsFormat
from .argument import Argument
from .option import Option
from .command_name import CommandName
from .command_option import CommandOption


class ArgsFormatBuilder(object):
    """
    A builder for ArgsFormat instances.
    """

    def __init__(self, base_format=None):  # type: (Optional[ArgsFormat]) -> None
        self._base_format = base_format
        self._command_names = []
        self._command_options = OrderedDict()
        self._command_options_by_short_name = OrderedDict()
        self._arguments = OrderedDict()
        self._options = OrderedDict()
        self._options_by_short_name = OrderedDict()
        self._has_multi_valued_arg = False
        self._hash_optional_arg = False

    @property
    def base_format(self):  # type: () -> Optional[ArgsFormat]
        return self._base_format

    def set_command_names(
        self, *command_names
    ):  # type: (Tuple[CommandName]) -> ArgsFormatBuilder
        self._command_names = []

        self.add_command_names(*command_names)

        return self

    def add_command_names(
        self, *command_names
    ):  # type: (Tuple[CommandName]) -> ArgsFormatBuilder
        for command_name in command_names:
            self.add_command_name(command_name)

        return self

    def add_command_name(
        self, command_name
    ):  # type: (CommandName) -> ArgsFormatBuilder
        self._command_names.append(command_name)

        return self

    def has_command_names(self, include_base=True):  # type: (bool) -> bool
        if self._command_names:
            return True

        if include_base and self._base_format:
            return self._base_format.has_command_names()

        return False

    def get_command_names(self, include_base=True):  # type: (bool) -> List[CommandName]
        command_names = self._command_names

        if include_base and self._base_format:
            command_names = self._base_format.get_command_names() + command_names

        return command_names

    def set_command_options(
        self, *command_options
    ):  # type: (Tuple[CommandOption]) -> ArgsFormatBuilder
        self._command_options = {}
        self._command_options_by_short_name = {}

        self.add_command_options(*command_options)

    def add_command_options(
        self, *command_options
    ):  # type: (Tuple[CommandOption]) -> ArgsFormatBuilder
        for command_option in command_options:
            self.add_command_option(command_option)

    def add_command_option(
        self, command_option
    ):  # type: (CommandOption) -> ArgsFormatBuilder
        long_name = command_option.long_name
        short_name = command_option.short_name
        long_aliases = command_option.long_aliases
        short_aliases = command_option.short_aliases

        if self.has_option(long_name) or self.has_command_option(long_name):
            raise CannotAddOptionException.already_exists(long_name)

        for long_alias in long_aliases:
            if self.has_option(long_alias) or self.has_command_option(long_alias):
                raise CannotAddOptionException.already_exists(long_alias)

        if self.has_option(short_name) or self.has_command_option(short_name):
            raise CannotAddOptionException.already_exists(short_name)

        for short_alias in short_aliases:
            if self.has_option(short_alias) or self.has_command_option(short_alias):
                raise CannotAddOptionException.already_exists(short_alias)

        self._command_options[long_name] = command_option

        if short_name:
            self._command_options_by_short_name[short_name] = command_option

        for long_alias in long_aliases:
            self._command_options[long_alias] = command_option

        for short_alias in short_aliases:
            self._command_options_by_short_name[short_alias] = command_option

        return self

    def has_command_option(self, name, include_base=True):  # type: (str, bool) -> bool
        if name in self._command_options or name in self._command_options_by_short_name:
            return True

        if include_base and self._base_format:
            return self._base_format.has_command_option(name)

        return False

    def has_command_options(self, include_base=True):  # type: (bool) -> bool
        if self._command_options:
            return True

        if include_base and self._base_format:
            return self._base_format.has_command_options()

        return False

    def get_command_option(
        self, name, include_base=True
    ):  # type: (str, bool) -> CommandOption
        if name in self._command_options:
            return self._command_options[name]

        if name in self._command_options_by_short_name:
            return self._command_options_by_short_name[name]

        if include_base and self._base_format:
            return self._base_format.get_command_option(name)

        raise NoSuchOptionException(name)

    def get_command_options(
        self, include_base=True
    ):  # type: (bool) -> Iterable[CommandOption]
        command_options = list(self._command_options.values())

        if include_base and self._base_format:
            command_options += self._base_format.get_command_options()

        return command_options

    def set_arguments(
        self, *arguments
    ):  # type: (Iterable[Argument]) -> ArgsFormatBuilder
        self._arguments = {}
        self._has_multi_valued_arg = False
        self._hash_optional_arg = False

        self.add_arguments(*arguments)

        return self

    def add_arguments(
        self, *arguments
    ):  # type: (Iterable[Argument]) -> ArgsFormatBuilder
        for argument in arguments:
            self.add_argument(argument)

        return self

    def add_argument(self, argument):  # type: (Argument) -> ArgsFormatBuilder
        name = argument.name

        if self.has_argument(name):
            raise CannotAddArgumentException.already_exists(name)

        if self.has_multi_valued_argument():
            raise CannotAddArgumentException.cannot_add_after_multi_valued()

        if argument.is_required() and self.has_optional_argument():
            raise CannotAddArgumentException.cannot_add_required_after_optional()

        if argument.is_multi_valued():
            self._has_multi_valued_arg = True

        if argument.is_optional():
            self._hash_optional_arg = True

        self._arguments[name] = argument

        return self

    def has_argument(
        self, name, include_base=True
    ):  # type: (Union[str, int], bool) -> bool
        arguments = self.get_arguments(include_base)

        if isinstance(name, int):
            return name < len(arguments)

        return name in arguments

    def has_multi_valued_argument(self, include_base=True):  # type: (bool) -> bool
        if self._has_multi_valued_arg:
            return True

        if include_base and self._base_format:
            return self._base_format.has_multi_valued_argument()

        return False

    def has_optional_argument(self, include_base=True):  # type: (bool) -> bool
        if self._hash_optional_arg:
            return True

        if include_base and self._base_format:
            return self._base_format.has_optional_argument()

        return False

    def has_required_argument(self, include_base=True):  # type: (bool) -> bool
        if not self._hash_optional_arg and self._arguments:
            return True

        if include_base and self._base_format:
            return self._base_format.has_required_argument()

        return False

    def has_arguments(self, include_base=True):  # type: (bool) -> bool
        if self._arguments:
            return True

        if include_base and self._base_format:
            return self._base_format.has_arguments()

        return False

    def get_argument(
        self, name, include_base=True
    ):  # type: (Union[str, int], bool) -> Argument
        if isinstance(name, int):
            arguments = list(self.get_arguments(include_base).values())

            if name >= len(arguments):
                raise NoSuchArgumentException(name)
        else:
            arguments = self.get_arguments(include_base)

            if name not in arguments:
                raise NoSuchArgumentException(name)

        return arguments[name]

    def get_arguments(self, include_base=True):  # type: (bool) -> Dict[str, Argument]
        arguments = self._arguments.copy()

        if include_base and self._base_format:
            arguments.update(self._base_format.get_arguments())

        return arguments

    def set_options(self, *options):  # type: (Iterable[Option]) -> ArgsFormatBuilder
        self._options = {}
        self._options_by_short_name = {}

        self.add_options(*options)

    def add_options(self, *options):  # type: (Iterable[Option]) -> ArgsFormatBuilder
        for option in options:
            self.add_option(option)

    def add_option(self, option):  # type: (Option) -> ArgsFormatBuilder
        long_name = option.long_name
        short_name = option.short_name

        if self.has_option(long_name) or self.has_command_option(long_name):
            raise CannotAddOptionException.already_exists(long_name)

        if self.has_option(short_name) or self.has_command_option(short_name):
            raise CannotAddOptionException.already_exists(short_name)

        self._options[long_name] = option

        if short_name:
            self._options_by_short_name[short_name] = option

        return self

    def has_option(self, name, include_base=True):  # type: (str, bool) -> bool
        if name in self._options or name in self._options_by_short_name:
            return True

        if include_base and self._base_format:
            return self._base_format.has_option(name)

        return False

    def has_options(self, include_base=True):  # type: (bool) -> bool
        if self._options:
            return True

        if include_base and self._base_format:
            return self._base_format.has_options()

        return False

    def get_option(self, name, include_base=True):  # type: (str, bool) -> Option
        if name in self._options:
            return self._options[name]

        if name in self._options_by_short_name:
            return self._options_by_short_name[name]

        if include_base and self._base_format:
            return self._base_format.get_option(name)

        raise NoSuchOptionException(name)

    def get_options(self, include_base=True):  # type: (bool) -> Dict[str, Option]
        options = self._options.copy()

        if include_base and self._base_format:
            options.update(self._base_format.get_options())

        return options

    @property
    def format(self):  # type: () -> ArgsFormat
        return ArgsFormat(self, self._base_format)
